# SPDX-License-Identifier: MIT
# Authors: Benjamin Dodge
import jax
import jax.numpy as jnp
import jax.random as jr

from jax.extend import core
from jax.core import ShapedArray

from jax.interpreters import batching, mlir, ad
from jax.flatten_util import ravel_pytree

from functools import partial

rng = jr.key(0)

param_p = core.Primitive("param")
fixed_p = core.Primitive("fixed")
intermediate_p = core.Primitive("intermediate")


def param(shape=(), *, name=None, dtype=None):
    """
    Marker for a parameter in a model passed to `purify`.
    These can be initialized with `model.zeros` or `model.normal`.
    The `name` must be unique among parameters in the model, if `None` is provided the object id will be used.
    """
    x = jnp.zeros(shape, dtype)
    return param_p.bind(x, name=name)


def fixed(shape=(), *, name, dtype=None):
    """
    Marker for a fixed value in a model passed to `purify`.
    Fixed values can often just be closed over, but this allows them to be passed as an argument instead.
    The `name` is required and must be unique among fixed values in the model.
    """
    x = jnp.zeros(shape, dtype)
    return fixed_p.bind(x, name=name)


def intermediate(x, *, name):
    """
    Marker for an intermediate value to be returned by a model passed to `purify` when called with `model.intermediates`.
    The `name` is required and must be unique among intermediates in the model.
    """
    return intermediate_p.bind(x, name=name)


def trivialize(primitive):
    """
    Register an "identity" primitive which exists solely as a marker in the jaxpr for `purify`.
    This could be useful for users who want to define markers for their own custom version of `purify`.
    """

    def abstract_eval(x, name=None):
        return ShapedArray(x.shape, x.dtype)

    def impl(x, name=None):
        return x

    def batch(args, batch_dims, name=None):
        return primitive.bind(args[0], name=name), batch_dims[0]

    def lowering(ctx, x, name=None):
        return [x]

    def transpose(xt, x, name=None):
        return xt,

    def jvp(primals, tangents, name=None):
        return primitive.bind(primals[0], name=name), tangents[0]

    primitive.def_impl(impl)
    primitive.def_abstract_eval(abstract_eval)
    batching.primitive_batchers[primitive] = batch
    mlir.register_lowering(primitive, lowering)
    ad.primitive_jvps[primitive] = jvp
    ad.primitive_transposes[primitive] = transpose


trivialize(param_p)
trivialize(fixed_p)
trivialize(intermediate_p)


def _eqn_name(eqn):
    name = eqn.params["name"]
    if name is None:
        name = str(id(eqn.outvars[0]))
    return name

def _set_eqn(dictionary, eqn, val):
    name = _eqn_name(eqn)
    if name in dictionary:
        raise ValueError(f"Duplicate name: {name}")
    dictionary[name] = val

def _safe_map(func, *iterables):
    for iterable in iterables:
        if len(iterable) != len(iterables[0]):
            raise ValueError("Iterables must have the same length.")
    return list(map(func, *iterables))

def purify(model=None, *, ravel=False, jit=False):
    """
    Transform a JAX function which takes no parameters into function which accepts parameters, optionally fixed data, and can return intermediate values.
    These behaviors are controlled by the use of `param`, `fixed`, and `intermediate` primitives in the original function. See the readme and examples folder for more usage details.

    Args:
        `model`: a JAX function which takes no parameters. Defaults to `None` in which case a partial function is returned which can be used as a decorator.
        `ravel`: if `True`, parameters will be flattened into a single 1d vector, also see `unravel` below.
        `jit`: if `True` the function will be automatically jitted.

    Returns:
        A pure function which should be called as `model(params, **fixed)` and has the additional attributes listed below. 

    Attributes:
        `shapes()`: returns a dictionary of parameter shapes and dtypes as `jax.ShapeDtypeStruct`.
        `zeros()`: returns a dictionary of parameters initialized to zero (or a 1d array if `ravel=True`).
        `normal(rng)`: returns a dictionary of parameters initialized with unit normal values (or a 1d array if `ravel=True`).
        `fixed_shapes()`: returns a dictionary of fixed value shapes and dtypes as `jax.ShapeDtypeStruct`.
        `intermediates(params, **fixed)`: returns a dictionary of intermediate values computed during evaluation.
        `unravel`: if `ravel=True`, unravels the raveled parameter vector back into the original parameter structure.
    """

    if model is None:
        return partial(purify, ravel=ravel)

    jaxpr = jax.make_jaxpr(model)()

    # Extract parameter structure for raveling
    unravel = None
    if ravel:
        zero_params = {}
        for eqn in jaxpr.jaxpr.eqns:
            if eqn.primitive is param_p:
                name = _eqn_name(eqn)
                if name in zero_params:
                    raise ValueError(f"Duplicate parameter: {name}")
                zero_params[name] = jnp.zeros_like(eqn.outvars[0].aval)
        _, unravel = ravel_pytree(zero_params)

    # Parameter structure function
    def shapes():
        """Return a dictionary of `param` arrays as `jax.ShapeDtypeStruct`."""
        params = {}
        for eqn in jaxpr.jaxpr.eqns:
            if eqn.primitive is param_p:
                _set_eqn(params, eqn, jax.ShapeDtypeStruct(eqn.outvars[0].aval.shape, eqn.outvars[0].aval.dtype))
                # outvar = eqn.outvars[0]
                # params[_eqn_name(eqn)] = jax.ShapeDtypeStruct(outvar.aval.shape, outvar.aval.dtype)
        if ravel:
            total_size = sum(param.size for param in params.values())
            cast_type = jnp.result_type(*[param.dtype for param in params.values()])
            return jax.ShapeDtypeStruct((total_size,), cast_type)
        return params

    # Zero initialization function
    def zeros():
        """Return a dictionary of `param` arrays initialized to zero."""
        params = {}
        for eqn in jaxpr.jaxpr.eqns:
            if eqn.primitive is param_p:
                _set_eqn(params, eqn, jnp.zeros_like(eqn.outvars[0].aval))
        if ravel:
            params, _ = ravel_pytree(params)
        return params

    # Random initialization function
    def normal(rng):
        """Return a dictionary of `param` arrays initialized with unit normal values."""
        params = {}
        for eqn in jaxpr.jaxpr.eqns:
            if eqn.primitive is param_p:
                rng, subkey = jr.split(rng)
                _set_eqn(params, eqn, jr.normal(subkey, eqn.outvars[0].aval.shape, eqn.outvars[0].aval.dtype))
        if ravel:
            params, _ = ravel_pytree(params)
        return params

    # Function to make fixed values full of zeros
    def fixed_shapes():
        """Return a dictionary of `fixed` arrays as `jax.ShapeDtypeStruct`. These should be passed as kwargs to model evaluation."""
        fixed_vals = {}
        for eqn in jaxpr.jaxpr.eqns:
            if eqn.primitive is fixed_p:
                _set_eqn(fixed_vals, eqn, jax.ShapeDtypeStruct(eqn.outvars[0].aval.shape, eqn.outvars[0].aval.dtype))
        return fixed_vals

    # Pure forward pass
    def forward(params, **fixed):
        """Evaluate model with params and (optionally) fixed values in the structure returned by `zeros`/`normal` and `fixed`, respectively."""
        if unravel:
            params = unravel(params)
        env = {}

        def read(v):
            return v.val if type(v) is core.Literal else env[v]

        def write(v, val):
            env[v] = val

        # Bind args and consts to environment
        _safe_map(write, jaxpr.constvars, jaxpr.consts)

        # Loop through equations and evaluate
        for eqn in jaxpr.jaxpr.eqns:
            # Load parameter or bind primitive
            if eqn.primitive is param_p:
                if _eqn_name(eqn) not in params:
                    raise ValueError(f"Model has parameter '{_eqn_name(eqn)}' but no value was provided.")
                outvals = params[_eqn_name(eqn)]
            elif eqn.primitive is fixed_p:
                if _eqn_name(eqn) not in fixed:
                    raise ValueError(f"Model has fixed value '{_eqn_name(eqn)}' but no value was provided.")
                outvals = fixed[_eqn_name(eqn)]
            else:
                invals = _safe_map(read, eqn.invars)
                result = eqn.primitive.get_bind_params(eqn.params)
                if isinstance(result, dict):  # JAX >= 0.9.2
                    outvals = eqn.primitive.bind(*invals, **result)
                else:  # JAX < 0.9.2
                    subfuns, bind_params = result
                    outvals = eqn.primitive.bind(*subfuns, *invals, **bind_params)
            # Save to context
            if not eqn.primitive.multiple_results:
                outvals = [outvals]
            _safe_map(write, eqn.outvars, outvals)
        # Return output
        out = _safe_map(read, jaxpr.outvars)
        return out[0] if len(out) == 1 else out

    # Forward pass returning intermediates
    def intermediates(params, **fixed):
        """Evaluate model and return intermediate values with params and (optionally) fixed values in the structure returned by `zeros`/`normal` and `fixed`, respectively."""
        if unravel:
            params = unravel(params)
        env = {}

        def read(v):
            return v.val if type(v) is core.Literal else env[v]

        def write(v, val):
            env[v] = val

        # Bind args and consts to environment
        _safe_map(write, jaxpr.constvars, jaxpr.consts)

        # Loop through equations and evaluate
        intermediate_outs = {}
        for eqn in jaxpr.jaxpr.eqns:
            # Load parameter or bind primitive
            if eqn.primitive is param_p:
                if _eqn_name(eqn) not in params:
                    raise ValueError(f"Model has parameter {_eqn_name(eqn)} but no value was provided.")
                outvals = params[_eqn_name(eqn)]
            elif eqn.primitive is fixed_p:
                if _eqn_name(eqn) not in fixed:
                    raise ValueError(f"Model has fixed parameter {_eqn_name(eqn)} but no value was provided.")
                outvals = fixed[_eqn_name(eqn)]
            else:
                invals = _safe_map(read, eqn.invars)
                result = eqn.primitive.get_bind_params(eqn.params)
                if isinstance(result, dict):  # JAX >= 0.9.2
                    outvals = eqn.primitive.bind(*invals, **result)
                else:  # JAX < 0.9.2
                    subfuns, bind_params = result
                    outvals = eqn.primitive.bind(*subfuns, *invals, **bind_params)
            # Save to context
            if not eqn.primitive.multiple_results:
                outvals = [outvals]
            _safe_map(write, eqn.outvars, outvals)
            # If intermediate, save to output dict
            if eqn.primitive is intermediate_p:
                intermediate_outs[_eqn_name(eqn)] = outvals[0] if len(outvals) == 1 else outvals

        return intermediate_outs

    forward.shapes = shapes
    forward.normal = normal
    forward.zeros = zeros
    forward.fixed_shapes = fixed_shapes
    forward.intermediates = jax.jit(intermediates) if jit else intermediates
    forward.unravel = unravel

    return jax.jit(forward) if jit else forward



import jax
import jax.numpy as jnp
import jax.random as jr

from .purify import param, intermediate


def normal_variable(name, mean=0.0, sigma=1.0):
    """Shortcut to create a scalar normal random variable which is both a parameter and intermediate value. Common for learned hyperparameters."""
    return intermediate(normal(param(name=name), mean=mean, sigma=sigma), name=name)


def log_normal_variable(name, mean=1.0, sigma=1.0):
    """Shortcut to create a scalar log-normal random variable which is both a parameter and intermediate value. Common for learned hyperparameters."""
    return intermediate(log_normal(param(name=name), mean=mean, sigma=sigma), name=name)


def uniform_variable(name, low=0.0, high=1.0):
    """Shortcut to create a scalar uniform random variable which is both a parameter and intermediate value. Common for learned hyperparameters."""
    return intermediate(uniform(param(name=name), low=low, high=high), name=name)


def normal(x, *, mean, sigma):
    """Transform a standard normal variable `x` into a normal variable with specified mean and standard deviation."""
    return mean + sigma * x


def log_normal(x, *, mean, sigma):
    """Transform a standard normal variable `x` into a log-normal variable with specified mean and standard deviation. Note that the mean must be positive."""
    mu = jnp.log(mean**2 / jnp.sqrt(mean**2 + sigma**2))
    sigma = jnp.sqrt(jnp.log(1 + sigma**2 / mean**2))
    return jnp.exp(mu + x * sigma)


def uniform(x, *, low=0.0, high=1.0):
    """Transform a standard normal variable `x` into a uniform variable with specified lower and upper bounds."""
    uniform = (1 + jax.lax.erf(x / jnp.sqrt(2))) / 2
    return low + (high - low) * uniform


def zeros_like(x):
    """Fill a pytree with zeros."""
    leaves, treedef = jax.tree_util.tree_flatten(x)
    zero_leaves = [jnp.zeros_like(leaf) for leaf in leaves]
    return jax.tree_util.tree_unflatten(treedef, zero_leaves)


def normal_like(rng, x):
    """Fill a pytree with unit normal values."""
    leaves, treedef = jax.tree_util.tree_flatten(x)
    normal_leaves = [jr.normal(rng, leaf.shape, leaf.dtype) for leaf in leaves]
    return jax.tree_util.tree_unflatten(treedef, normal_leaves)



import jax
import jax.numpy as jnp
import jax.random as jr

import jaxpurify as jp
from jaxpurify import purify

rng = jr.key(0)

def test_params():
    @purify
    def model():
        x = jp.param(3, name="x")
        return x**2

    params = model.normal(rng)
    result = model(params)

    assert params["x"].shape == (3,)
    assert jnp.allclose(result, params["x"]**2)

def test_shapes():
    @purify
    def model():
        x = jp.param((4,3), name="x")
        y = jp.param(2, name="y")

    shapes = model.shapes()
    zeros = model.zeros()
    params = model.normal(rng)

    assert shapes["x"].shape == (4, 3)
    assert shapes["y"].shape == (2,)
    assert zeros["x"].shape == (4, 3)
    assert zeros["y"].shape == (2,)
    assert params["x"].shape == (4, 3)
    assert params["y"].shape == (2,)

def test_ravel():
    @purify(ravel=True)
    def model():
        x = jp.param((4,3), name="x")
        y = jp.param(2, name="y")

    shapes = model.shapes()
    zeros = model.zeros()
    params = model.normal(rng)

    unraveled_params = model.unravel(params)

    assert shapes.shape == (14,)
    assert zeros.shape == (14,)
    assert params.shape == (14,)
    assert unraveled_params["x"].shape == (4, 3)
    assert unraveled_params["y"].shape == (2,)

def test_fixed():
    @purify
    def model():
        x = jp.param(3, name="x")
        b = jp.fixed((2,3), name="b")
        return x + b

    params = model.normal(rng)
    fixed = model.fixed_shapes()
    result = model(params, b=1.0)

    assert fixed["b"].shape == (2, 3)
    assert jnp.allclose(result, params["x"] + 1.0)

def test_intermediates():
    @purify
    def model():
        x = jp.param(3, name="x")
        y = jp.intermediate(x**2, name="y")
        return y + 1

    params = model.normal(rng)
    intermediates = model.intermediates(params)
    assert jnp.allclose(intermediates["y"], params["x"]**2)

def test_with_custom_jvp():
    @purify
    def model():
        return jax.nn.relu(jp.param(3, name="x"))

    params = model.normal(rng)
    results = model(params)
    assert jnp.allclose(results, jax.nn.relu(params["x"]))

def test_higher_order_primitives():
    def f(x):
        a = jp.param()
        return jnp.sin(a * x)

    @purify
    def model():
        x = jp.param(3, name="x")
        y = jax.jit(jax.vmap(jax.grad(jax.grad(f))))(x)
        return jnp.sum(y)

    def model_explicit(x):
        y = jax.jit(jax.vmap(jax.grad(jax.grad(f))))(x)
        return jnp.sum(y)

    params = model.normal(rng)
    result = model(params)
    result_explicit = model_explicit(params['x'])
    assert jnp.allclose(result, result_explicit)

    result_grad = jax.grad(model)(params)
    result_grad_explicit = jax.grad(model_explicit)(params['x'])
    assert jnp.allclose(result_grad['x'], result_grad_explicit)

    


"""
Demonstrate how to use jaxpurify to build neural networks.
This is not the targeted use case but the code turns out to be quite elegant.
One major downside is that neural networks are often trees with all arrays trainable
and so it is more natural to store the model as a callable PyTree, similar to Equinox or Flax.

"""

# %%
import jax
import jax.numpy as jnp
import jax.random as jr
from functools import partial

import jaxpurify as jp
from jaxpurify import purify

rng = jr.key(0)


# %% The simplest approach, unnamed parameters
def linear(x, n_in, n_out):
    w = jp.param((n_out, n_in))
    b = jp.param(n_out)
    return w @ x + b

@purify
def model():
    x = jp.fixed(16, name="x")
    x = linear(x, 16, 8)
    x = jax.nn.relu(x)
    x = linear(x, 8, 4)
    x = jax.nn.softmax(x)
    return x

print(model.shapes())
print(model(model.normal(rng), x=jnp.ones(16)))

# %% An alternative pattern for reference, now also with unique parameter names
def linear(n_in, n_out, name):
    def apply(x):
        w = jp.param((n_out, n_in), name=f"{name}_w")
        b = jp.param(n_out, name=f"{name}_b")
        return w @ x + b
    return apply

@purify
def model():
    layers = [
        linear(16, 8, name="linear_1"),
        jax.nn.relu,
        linear(8, 4, name="linear_2"),
        jax.nn.softmax,
    ]
    x = jp.fixed(16, name="x")
    for layer in layers: x = layer(x)
    return x

print(model.shapes())

# %% Factory function for building a dense network
def dense_network(dims):
    def model(x):
        n = len(dims) - 1
        for i in range(len(dims) - 1):
            if i > 0: x = jax.nn.relu(x)
            x = linear(dims[i], dims[i+1], name=f"linear_{i+1}")(x)
        return x
    return model

@purify
def model():
    x = jp.fixed(16, name="x")
    x = dense_network([16, 8, 4])(x)
    return jax.nn.softmax(x)

print(model.shapes())
# %%



"""
Infer a field from noisy sparse observations using a Gaussian process prior.
The measurement is linear and the noise is Gaussian so the posterior has an analytic solution.
We can approximately solve the required matrix equations using conjugate gradient iteration, avoiding expensive dense matrices.

Specifically, given a model d = R s + n where n ~ N(0, N) is noise and s ~ N(0, I) is unit normal a priori ("whitened" parameters),
we can solve for posterior mean m and mean-subtracted posterior samples z using

    D = (I + R^T N^-1 R)^-1
    m = D (R_T N^-1 d)
    z = D (R_T N^-1 q + p)
    
where q ~ N(0, I) and p ~ N(0, I) are independent noise samples of data and parameter dimension, respectively.
A few lines of math show that <z z^T> = (I + R^T N^-1 R)^-1, as desired.

"""

# %%
import jax
import jax.numpy as jnp
import jax.random as jr
from jax.scipy.sparse.linalg import cg
from functools import partial
import matplotlib.pyplot as plt

import jaxpurify as jp
from jaxpurify import purify

rng = jr.key(0)


# %% Define model
def field(n, pad=10):
    xi = jp.param(n + 2*pad, name="xi")
    k = jnp.fft.fftfreq(n + 2*pad, d=1.0/n)
    p = jnp.power(jnp.abs(k), -3.0).at[0].set(0.0)
    f = jnp.fft.fft(jnp.sqrt(p) * xi)
    f = (f.real - f.imag)
    return f[pad:n+pad]

@purify(ravel=True)
def model():
    pad = 10
    x = jp.fixed(1000, name="x")
    y = jp.intermediate(field(x.shape[0], pad=pad), name="y")
    x_obs = jp.fixed(100, name="x_obs")
    y_obs = jnp.interp(x_obs, x, y)
    return y_obs


# %% Simulate data
rng, k1, k2, k3 = jr.split(rng, 4)
x = jnp.linspace(0, 20, 1000)
x_obs = jr.uniform(k1, 100, minval=x.min(), maxval=x.max())
y_true = jnp.sin(x_obs)
noise_std = jnp.sqrt(jnp.abs(y_true))/2
noise = noise_std * jr.normal(k2, x_obs.shape)
y_obs = y_true + noise

# %% Solve for posterior mean using CG
R = partial(model, x=x, x_obs=x_obs)
R_T = lambda d: jax.linear_transpose(R, model.shapes())(d)[0]
N_inv = lambda x: x / noise_std**2
M = lambda x: R_T(N_inv(R(x))) + x
mean = cg(M, R_T(N_inv(y_obs)), x0=model.zeros(), maxiter=100)[0]
field_mean = model.intermediates(mean, x=x, x_obs=x_obs)["y"]

# %% Solve for posterior samples using CG and sampling trick
n_samples = 5
rng, k1, k2 = jr.split(rng, 3)
data_noise = jr.normal(k1, (n_samples, *y_obs.shape))
prior_noise = jr.normal(k2, (n_samples, *mean.shape))
samples = mean + jax.vmap(partial(cg, M, maxiter=100))(prior_noise + jax.vmap(R_T)(data_noise), prior_noise)[0]
field_samples = jax.vmap(partial(model.intermediates, x=x, x_obs=x_obs))(samples)["y"]

# %% Plot results!
plt.plot(x, jnp.sin(x), c='k', alpha=0.5)
plt.errorbar(x_obs, y_obs, yerr=noise_std, fmt="o", c='k', alpha=0.5, ms=2, lw=1)
plt.plot(x, field_mean, c='C0')
plt.plot(x, field_samples.T, alpha=0.5, c='C0')
plt.show()
# %%



from .purify import purify, param, fixed, intermediate
from .models import (
    normal_variable,
    log_normal_variable,
    uniform_variable,
    normal,
    log_normal,
    uniform,
    zeros_like,
    normal_like,
)

__all__ = [
    "purify",
    "param",
    "fixed",
    "intermediate",
    "normal_variable",
    "log_normal_variable",
    "uniform_variable",
    "normal",
    "log_normal",
    "uniform",
    "zeros_like",
    "normal_like",
]



"""
Explain the basic features of jaxpurify:
 - parameter management with `shapes`, `zeros`, and `normal`
 - modularity with parameterized model components
 - fixed array inputs with `fixed`
 - intermediate return values with `intermediate`
 - flat vector parameters with `ravel=True` and `unravel`
- shortcut for `normal`, `uniform`, and `log_normal` variables

"""

# %%
import jax
import jax.numpy as jnp
import jax.random as jr
from functools import partial

import jaxpurify as jp
from jaxpurify import purify

rng = jr.key(0)

# %% Basic model with parameter shapes, zeros, and normal
@purify
def model():
    x = jp.param(3, name="x")
    return x**2

param_shapes = model.shapes()
zero_params = model.zeros()
params = model.normal(rng)
result = model(params)

print("Param shapes:", param_shapes)
print("Zero params:", zero_params)
print("Normal Params:", params)
print("Result:", result)

# %% Parameters can be defined in any function called by the model
def width():
    return jp.param(name="w")

def height():
    return jp.param(name="h")

@purify
def model():
    return jnp.sqrt(width()**2 + height()**2)

params = model.normal(rng)
result = model(params)
print("Params:", params)
print("Result:", result)

# %% Fixed arrays can be closed over or passed as keyword arguments
@purify
def model():
    a = jnp.array(1.0)
    b = jp.fixed(name="b")
    return a + b

fixed_shapes = model.fixed_shapes()
result = model(None, b=2.0)
print("Fixed shapes:", fixed_shapes)
print("Result:", result)

# %% Intermediate values can be returned and accessed by name
@purify
def model():
    x = jp.param(3, name="x")
    y = jp.intermediate(x**2, name="y")
    return jnp.sqrt(jnp.sum(y))

params = model.normal(rng)
intermediates = model.intermediates(params)
result = model(params)
print("Intermediates:", intermediates)
print("Result:", result)

# %% Parameters can be flattened into a single 1d array
@purify(ravel=True, jit=True)
def model():
    x = jp.param(2)
    y = jp.param((2, 2))
    return jnp.sum(x * y)

flat_params = model.normal(rng)
unraveled_params = model.unravel(flat_params)
result = model(flat_params)

print("Flat params:", flat_params)
print("Unraveled params:", unraveled_params)
print("Result:", result)

# %% Shortcuts for common variable types, the "variables" are automatically intermediates
@purify
def model():
    a = jp.normal_variable("a", mean=1.0, sigma=2.0)
    b = jp.log_normal_variable("b", mean=1.0, sigma=0.5)
    c = jp.uniform_variable("c", low=-1.0, high=1.0)

    A = jp.normal(jp.param((3, 3), name="A"), mean=1.0, sigma=2.0)
    B = jp.log_normal(jp.param((3, 3), name="B"), mean=1.0, sigma=0.5)
    C = jp.uniform(jp.param((3, 3), name="C"), low=-1.0, high=1.0)
    return a, b, c

params = model.normal(rng)
result = model(params)
intermediates = model.intermediates(params)
print("Params:", params)
print("Result:", result)
print("Intermediates:", intermediates)
# %%



"""
Demonstrate how to build a deep, complex model with multiple reusable components.
In this case, all parameters are defined in a centralized "orchestrator" function,
and passed as needed to various components. It is just as easy to define parameters
in the components themselves as shown in other examples, depending on what makes the
most sense for the application.
"""

# %%
import jax
import jax.numpy as jnp
import jax.random as jr
from functools import partial

import jaxpurify as jp
from jaxpurify import purify

rng = jr.key(0)


# %% Define model
def chop(vegetable, slices):
    return (vegetable / slices) * jnp.ones(slices)

def boil(ingredients, *, time):
    kernel = jnp.exp(-(jnp.linspace(-60/(time+1), 60/(time+1), 15) ** 2))
    return jnp.convolve(ingredients, kernel, mode="same")

def sautee(ingredients, *, time):
    scale = jnp.exp(time / 5) / jnp.sqrt(ingredients.size)
    return scale * jnp.cumsum(ingredients)

def prepare_aromatics(onions, carrots, celery, *, time_onions, time_together):
    chop_multiple = jax.vmap(chop, in_axes=(0, None))
    onions = chop_multiple(onions, 30).flatten()
    carrots = chop_multiple(carrots, 5).flatten()
    celery = chop_multiple(celery, 5).flatten()
    onions = sautee(onions, time=time_onions)
    aromatics = sautee(jnp.concatenate([onions, carrots, celery]), time=time_together)
    return aromatics

def make_soup(lentils, aromatics, seasoning, lemon, *, boil_time):
    soup = boil(seasoning * jnp.concatenate([lentils, aromatics]), time=boil_time)
    soup += lemon
    return soup

def eat(food):
    return jnp.sum(food**2)

@purify
def dinner():
    carrots = jnp.cos(jp.param(5, name="carrots"))
    celery = jnp.sin(jp.param(5, name="celery"))
    onions = jnp.tanh(jp.param(2, name="onions"))
    aromatics = prepare_aromatics(
        onions, carrots, celery,
        time_onions=jp.log_normal_variable("onion_time", mean=10, sigma=1),
        time_together=jp.log_normal_variable("aromatics_time", mean=3, sigma=1),
    )
    aromatics = jp.intermediate(aromatics, name="aromatics")

    seasoning = sum([
        jp.log_normal_variable("salt", mean=10, sigma=2),
        jp.log_normal_variable("thyme", mean=3, sigma=1),
        jp.uniform_variable("bay_leaf", low=2, high=3),
    ])

    lentils = jp.log_normal(jp.param(300, name="lentils"), mean=1.0, sigma=0.1)
    lentils = jax.jvp(jnp.sin, (lentils,), (lentils,))[1]

    soup = make_soup(
        lentils, aromatics, seasoning,
        jp.log_normal_variable("lemon", mean=0.4, sigma=0.1),
        boil_time=jp.uniform_variable("boil_time", low=20, high=30),
    )
    soup = jp.intermediate(soup, name="soup")

    return eat(soup)

# %% Run model
params = dinner.normal(rng)
result = dinner(params)
intermediates = dinner.intermediates(params)
lots_of_soup = jax.vmap(dinner.normal)(jr.split(rng, 10))
grads = jax.jit(jax.vmap(jax.grad(dinner)))(lots_of_soup)

# %%
